/*
 WSE, TSD US Naval Academy 
 Editor     : Joe Bradshaw, based on Tong Hui from DFRobot, based on Lauren from DFRobot v1.0 code

 Product name: Wireless Gamepad v2.2 for Arduino
 Product SKU : DFR0182
 Code Version: 2.1

 Description:
 The sketch for using the gamepad and print the button state and the analog values of the gamepad
 Enable the vibration function
   Send 'v' via the Arduino Serial monitor to enable the vibration
   Send 's' via the Arduino Serial monitor to stop it
   Send '?' to read the Xbee addresses
  20141017 - Swapped JSLX and JSLY, inverted X axis
       added sleep timer for warning and sleep mode
  20141021 - added debug 'd' command for countdown timer and revision 'r' command
*/

#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <Sleep_n0m1.h>

#define WARNING_SEC  3000  // (3000) 5 min = 300 sec
#define SLEEP_SEC  3120    // (3120) 12 sec warning to sleep
#define HYST  2

int buttonState[17];
int testDummy;
unsigned int buttonWordTemp;
unsigned int buttonWord;
int joystick_LX;
int joystick_LY;
int joystick_RX;
int joystick_RY;
String Buttons[16] = {    // Buttons Names
  "LJB","RJB","SELECT","START","UP","LEFT","DOWN","RIGHT","1","4","2","3","RZ1","RZ2","LZ1","LZ2"};
int inputCommand = 0;
int mode = 0;

unsigned int buttonWord_old;
int joystick_LX_old;
int joystick_LY_old;
int joystick_RX_old;
int joystick_RY_old;

//Pin Definitions
int vibMotPin = 2;  //output for bibration motors
int rxLED = 17;    // Receive LED

unsigned long timer = 0;
unsigned long sleepTimerCount = 0;
boolean sleepState = false;

int sleepFlag = 0;
Sleep sleep;
unsigned long sleepTime; //how long you want the arduino to sleep
int js_state=0;

int debugFlag = 0;
//-------- Setup -------------------------------
void setup()
{
  delay(200);
  sleepTime = 26800; //set sleep time in ms, max sleep time is 49.7 days

  Serial.begin(115200);  //Init the Serial baudrate
  Serial1.begin(9600);  //Init the Serial1 port to enable the xbee wireless communication
  InitIO();             //Initialize the inputs/outputs and the buffers 
 
  delay(20);
  DataUpdate();  //read the buttons and the joysticks data

  //initialize history states as current for first iteration
  buttonWord_old = buttonWord;
  joystick_LX_old = joystick_LX;
  joystick_LY_old = joystick_LY;
  joystick_RX_old = joystick_RX;
  joystick_RY_old = joystick_RY;

// Check MODE
  //if((Buttons[12]==0) && (Buttons[14] == 0)) //if LZ1 and RZ1 are pushed on power up, query mode
  if(buttonWord == 0xFFF5)
    mode=1;
    
  Serial.print(buttonWord);
  delay(200);
  //setup timer interrupt
  cli();//stop interrupts
  
//set timer1 interrupt at 10Hz
  TCCR1A = 0;// set entire TCCR1A register to 0
  TCCR1B = 0;// same for TCCR1B
  TCNT1  = 0;//initialize counter value to 0
  // set compare match register for 10hz increments (1562 with prescaler of 1024)
  OCR1A = 1562;// = (16*10^6) / (10*1024) - 1 //(must be <65536)
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS12 and CS10 bits for 1024 prescaler
  TCCR1B |= (1 << CS12) | (1 << CS10);  
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);

  sei();//allow interrupts
}

void loop(){  
  if(mode == 0)
  {
    if(millis() - timer > 50){  // manage the updating freq of all the controlling information
    
        //FOR TESTING
      if(debugFlag==1){
        Serial.print(millis());
        Serial.print("  ");  
        
        Serial.print(sleepTimerCount);
        Serial.print("  ");
        
        Serial1.print(millis());
        Serial1.print("  ");  
        
        Serial1.print(sleepTimerCount);
        Serial1.print("  ");  
      }
  
      DataUpdate();  //read the buttons and the joysticks data
      printData();   //print the data and states      
      timer = millis(); 
    }
    
    if(Serial1.available()){
      char input = Serial1.read();
      
      switch(input){
        case '?':
          Serial1.println("v - vibrate motor");
          Serial1.println("s - stop motor");
          Serial1.println("d - debug mode");
          Serial1.println("x - xbee query addresses");
          Serial1.println("r - revision of software");
          delay(3000);          
          inputCommand = input;          
          break;        
        
        case 'v':
          Serial1.println("Vibration");
          inputCommand = input;
          digitalWrite(vibMotPin,HIGH);
          break;
        
        case 's':
          Serial1.println("Stop");
          inputCommand = input;
          digitalWrite(vibMotPin,LOW);
          break;
          
        case 'x':
          read_xbee();
          inputCommand = input;            
          break;          
        
        case 'd':
          debugFlag ^= 1;
          inputCommand = input;
          break;
          
        case 'r':
          Serial1.println("DF Robot Gamepad - WSE ver 20141021");
          delay(3000);
          inputCommand = input;
          break;          
            
        default:
          break;
      }
    }
    
    if(Serial.available()){
      char inputc = Serial.read();
      
      switch(inputc){
        case '?':
          Serial.println("v - vibrate motor");
          Serial.println("s - stop motor");
          Serial.println("d - debug mode");
          Serial.println("x - xbee query addresses");
          Serial.println("r - revision of software");
          delay(3000);          
          inputCommand = inputc;
          break;        
        
        case 'v':
          inputCommand = inputc;
          digitalWrite(vibMotPin,HIGH);
          break;
        
        case 's':
          inputCommand = inputc;
          digitalWrite(vibMotPin,LOW);
          break;
          
        case 'x':
          read_xbee();
          inputCommand = inputc;            
          break;          
        
        case 'd':
          debugFlag ^= 1;
          inputCommand = inputc;
          break;
          
        case 'r':
          Serial.println("DF Robot Gamepad - WSE ver 20141021");
          delay(3000);
          inputCommand = inputc;
          break;           
            
        default:
          break;
      }           
    }
  }//mode 0
    else if(mode == 1){
      if(Serial1.available()){
      char input = Serial1.read();
      
        switch(input){
          case 0x0d:  //transmit data upon receiving carriage return          
                  //FOR TESTING
            if(debugFlag==1){
              Serial.print(millis());
              Serial.print("  ");  
              
              Serial.print(sleepTimerCount);
              Serial.print("  ");
              
              Serial1.print(millis());
              Serial1.print("  ");  
              
              Serial1.print(sleepTimerCount);
              Serial1.print("  ");  
            }
            DataUpdate();  //read the buttons and the joysticks data
            printData();   //print the datas and states
            break;
  
          case 'v':
            Serial1.println("Vibration");
            inputCommand = input;
            digitalWrite(vibMotPin,HIGH);
            break;
          
          case 's':
            Serial1.println("Stop");
            inputCommand = input;
            digitalWrite(vibMotPin,LOW);
            break;

          case 'x':
            inputCommand = input;
            read_xbee();            
            break;
            
          case 'd':
            debugFlag ^= 1;
            inputCommand = input;
            break;            

          case '?':
            Serial.println("v - vibrate motor");
            Serial.println("s - stop motor");
            Serial.println("d - debug mode");
            Serial.println("x - xbee query addresses");
            Serial.println("r - revision of software");
            delay(3000);
            inputCommand = input;
            break;             

          case 'r':
            Serial.println("DF Robot Gamepad - WSE ver 20141021");
            delay(3000);
            inputCommand = input;
            break; 
            
          default:
            break;
        }      
      }
    }//else mode == 1
    
    if(sleepFlag == 1){
      digitalWrite(vibMotPin,LOW);      
      digitalWrite(rxLED,HIGH);      
      Serial.print("Going to sleep");
      Serial1.print("Going to sleep");
      
      delay(1400);

      Serial.flush();  
      
      Serial1.print("+");  //put xbee in command mode
      delay(20);
      Serial1.print("+");  //put xbee in command mode
      delay(20);
      Serial1.print("+");  //put xbee in command mode
      delay(20);  
      
      char ack_str[10];
      Serial1.setTimeout(2000);
      if(Serial1.readBytes(ack_str, 3)){
          Serial.println(ack_str);  //transmit the read character out the debug serial port
      }
      else
        Serial.println("No ACK");
     
      Serial1.print("ATST 200\r");  //set the time before sleep to 200ms
      delay(200);
      Serial1.print("ATSP 26800\r");  //set the cyclic sleep time to 268ms
      delay(50);
      Serial1.print("ATSM 4\r");
      delay(50);
      Serial1.print("ATCN\r");
      
      cli();//stop interrupts
      delay(100);
      sleep.pwrDownMode(); //set sleep mode
      sleep.sleepDelay(sleepTime); //sleep for: sleepTime      
    }
}//loop()

//---- Methods --------------------------------
void read_xbee(){
  int strcnt=0;
  char ATDL_str[10];    //string to store xbee data
  char ATMY_str[10];
  char ack_str[10];
  
  for(int i=0;i<10;i++){
     ATDL_str[i] = 0; 
     ATMY_str[i] = 0;
     ack_str[i] = 0;
  }
  
  Serial.println("Entering command mode - Reading Xbee addresses");
  Serial1.println("Entering command mode - Reading Xbee addresses");
  delay(1400);
  Serial.flush();  
  
  Serial1.print("+");  //put xbee in command mode
  delay(20);
  Serial1.print("+");  //put xbee in command mode
  delay(20);
  Serial1.print("+");  //put xbee in command mode
  delay(20);  
  
  Serial1.setTimeout(2000);
  if(Serial1.readBytes(ack_str, 3)){
      Serial.println(ack_str);  //transmit the read character out the debug serial port
  }
  else
    Serial.println("No ACK");
 
  Serial1.print("ATDL\r");  //query xbee for Destination address

  Serial1.setTimeout(2000);
  if(Serial1.readBytes(ATDL_str, 5)){
      Serial.print("ATDL ");
      Serial.println(ATDL_str);  //transmit the read character out the debug serial port
  }
  else
    Serial.println("Error reading Destination address low");

  Serial1.print("ATMY\r");  //query xbee for My address
  
  Serial1.setTimeout(2000);
  if(Serial1.readBytes(ATMY_str, 5)){
      Serial.print("ATMY ");  //transmit the read character out the debug serial port
      Serial.println(ATMY_str);  //transmit the read character out the debug serial port
  }
  else
    Serial.println("Error reading My address");
    
  Serial1.print("ATCN\r");  //exit command mode
  
  Serial1.print("ATDL ");
  Serial1.println(ATDL_str);
  
  Serial1.print("ATMY ");
  Serial1.println(ATMY_str);
  
  delay(3000);  
}

void InitIO(){ 
  for(int i = 0; i < 16; i++){
    pinMode(i, INPUT);
  } 
  digitalWrite(vibMotPin,LOW);  // Stop shaking of the gamepad
  pinMode(vibMotPin,OUTPUT);
  digitalWrite(rxLED,HIGH);  // Receive LED indicator
  pinMode(rxLED,OUTPUT);
}

void DataUpdate(){
  buttonWordTemp = 0;  //zero buttonWordTemp
  
    //Read Left Joystick pushbutton (buttonWord MSB)
  if(analogRead(1) > 512){
      buttonState[0] = 1;
      buttonWordTemp += 1;
  }
  else
    buttonState[0] = 0;
  
  buttonWordTemp = buttonWordTemp << 1;
  if(analogRead(0) > 512){
      buttonState[1] = 1;
      buttonWordTemp += 1;
  }
  else
    buttonState[1] = 0;
  
  for(int i = 3; i < 17; i++){
    buttonWordTemp = buttonWordTemp << 1;
    if(digitalRead(i)){      
      buttonWordTemp += 1;
    }
    
    buttonState[i] = digitalRead(i);
  } 
  
  buttonWord = buttonWordTemp;
  joystick_RY = analogRead(2);
  joystick_RX = 1023 - analogRead(3);
  joystick_LY = analogRead(4);
  joystick_LX = 1023 - analogRead(5);  
}

int get_checksum(char *sentence)
{
  int checksum;
  int i;
  int MAX_CHARS = 100;  //define the MAX bumber of characters

  i=0;
  checksum = 0;

  if(sentence[i] == '$')  //Check for beginning '$' character
    i++;
  else
    return 0;  //exit with zero if invalid NMEA string
	
  //Iterate throught the string, terminate on '*', NULL, or character overflow
  while((sentence[i] != '*') && (sentence[i] != '\0') && (i < MAX_CHARS))
  {
    checksum ^= sentence[i];  //XOR next char to checksum
    i++;
  }

  if(i >= MAX_CHARS)  //exit with 0 if MAX_CHARS was reached
    return 0;

  return checksum;
}

void printData(){
  int checksum;
  char sent[100];
  char tempstr[50];
  
  //begin building the string
  sprintf(sent, "$JOYSTK,");      //header
  
  sprintf(tempstr, "%04X,", buttonWord);    
  strcat(sent, tempstr);    //and concatinate to sent

  sprintf(tempstr, "%d,", joystick_LX);  //add LXjoystick data
  strcat(sent, tempstr);    //and concatinate to sent
  
  sprintf(tempstr, "%d,", joystick_LY);  //add LYjoystick data
  strcat(sent, tempstr);    //and concatinate to sent
  
  sprintf(tempstr, "%d,", joystick_RX);  //add RXjoystick data
  strcat(sent, tempstr);    //and concatinate to sent

  sprintf(tempstr, "%d", joystick_RY);  //add RYjoystick data, No Comma
  strcat(sent, tempstr);    //and concatinate to sent  
  
  strcat(sent, "*");  //append an astrisk to the string
  
  checksum = get_checksum(sent);  //calculate the checksum of the NMEA string
  checksum &= 0xff;                //bit mask lower 8-bits
  sprintf(tempstr, "%02X\r\n", checksum);    //append ASCII checksum (HEX) to string
  strcat(sent, tempstr);    //and concatinate to sent    
  
  Serial.print(sent);    //transmit to the USBtoSerial programming port
  Serial1.print(sent);    //transmit to Xbee    
}//print data

ISR(TIMER1_COMPA_vect)
{
  if((buttonWord_old == buttonWord) &&       
      ((joystick_LX_old <= joystick_LX+HYST) && (joystick_LX_old >= joystick_LX-HYST)) &&
      ((joystick_LY_old <= joystick_LY+HYST) && (joystick_LY_old >= joystick_LY-HYST)) &&
      ((joystick_RX_old <= joystick_RX+HYST) && (joystick_RX_old >= joystick_RX-HYST)) &&
      ((joystick_RY_old <= joystick_RY+HYST) && (joystick_RY_old >= joystick_RY-HYST))){
      
      sleepTimerCount++;  //if no change the increment the sleep timer count variable
  }
  else{
    if(js_state == 1){
        digitalWrite(vibMotPin,LOW);  //turn off motor
        js_state = 0;
    }
    sleepTimerCount = 0;  //clear sleepTimerState if change occured
  }
  
  if((sleepTimerCount >= WARNING_SEC) && (sleepTimerCount <= SLEEP_SEC)){  // 3 min
    js_state = 1;
    if(sleepState == false){
        digitalWrite(rxLED,HIGH);
        digitalWrite(vibMotPin,LOW);  //turn off motor
        sleepState = true;
    }
    else{
        digitalWrite(rxLED,LOW);
        digitalWrite(vibMotPin,HIGH);
        sleepState = false;      
    }    
  }
  
   if(sleepTimerCount > SLEEP_SEC){
      js_state=2;
      digitalWrite(vibMotPin,LOW);
      sleepFlag = 1;  //indicate sleep mode
   }  
  
  //update history states
  buttonWord_old = buttonWord;
  joystick_LX_old = joystick_LX;
  joystick_LY_old = joystick_LY;
  joystick_RX_old = joystick_RX;
  joystick_RY_old = joystick_RY;
}
